/*******************************************************************************
 * Copyright (c) 2003, 2007 s IT Solutions AT Spardat GmbH .
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *     s IT Solutions AT Spardat GmbH - initial API and implementation
 *******************************************************************************/

package at.spardat.enterprise.fmt;

import java.util.ArrayList;
import java.util.BitSet;
import java.util.Iterator;
import java.util.List;

/**
 * Compound validator, which manages a chain of validators.
 * It delegates all validation checks to a the validators contained in its chain.
 * The input is accepted, if it is accepted by all enabled validators in the chain.
 * If an formatter is explicitly set for formatting, it is used. Otherwise the
 * first validator in the chain is used for formatting.
 * 
 *   @author gub
 */
public class CompoundValidator extends IFmt {
	// list containing all IFmts used for validation
	List chain = new ArrayList();
	// enabled/disabled switches for the IFmts in chain
	// each bit controls the IFmt with the corresponding index in chain
	BitSet switches = new BitSet();
	// IFmt used for formatting
	IFmt formatter;
	
	/**
	 * Creates a CompoundValidator with the given chain of validators.
	 * The chain must contain only IFmt objects.
	 * More validators can be added later on. 
	 */
	public CompoundValidator(List chain) {
		setChain(chain);
	}
	
	/**
	 * Creates a CompoundValidator with one validator in its chain.
	 * More validators can be added later on. 
	 */
	public CompoundValidator(IFmt validator) {
		addToChain(validator);
	}
	
	/**
	 * Creates a CompoundValidator with two validators in its chain.
	 * More validators can be added later on. 
	 */
	public CompoundValidator(IFmt validator1,IFmt validator2) {
		addToChain(validator1);
		addToChain(validator2);
	}
	
    /**
     * Transforms an internal encoding to an external.
     * If a formatter is explicitly set by {@link #setFormatter(IFmt)},
     * this formatter is used for formatting. Otherwise the first validatior
     * of the chain is used. 
     * 
     * @param internal the provided internal encoding. This String must satisfy the
     *         condition <tt>isLegalInternal()</tt>, otherwise the behaviour of this
     *         method is undefined.
     * @return the external representation. May be the empty string, but is never null.
     */     
	public String format(String internal) {
		return getFmtUsedForFormatting().format(internal);
	}

    /**
     * An <tt>IFmt</tt> is one way, if the external encoding may be generated out of
     * the internal, but not vice versa. A CompoundValidator is oneWay, if the
     * formatter is oneWay. (set explicitly or first of chain)
     * 
     * @return true, if <tt>format</tt> may be called, but <tt>parse</tt> not.
     */ 
	public boolean isOneWay() {
		return getFmtUsedForFormatting().isOneWay();
	}

    /**
     * Tries to transform an external encoding to an internal. Must not be called if
     * <tt>isOneWay()</tt> yields true.
     * Every enabled validator in the chain tries to parse the string. A AParseException is
     * thrown if any of the enabled validators in the chain fails. Disabled validators in
     * the chain are ignored. The result is the result of the used formatter.
     * 
     * @param external the external encoding
     * @return the internal encoding. May be the empty String, but never is null.
     * @exception AParseException if the external encoding cannot be successfully parsed.
     */
	public String parse(String external) throws AParseException {
		for(int i=0;i<chain.size();i++) {
			if(switches.get(i)) {
				// give everyone a chance to complain
				((IFmt)chain.get(i)).parse(external);
			}
		}
		// use representation of formatter
		String internal = getFmtUsedForFormatting().parse(external);
		checkMandatory(internal);
		return internal;
	}


    /**
     * Defines if this formatter is able to successfully format a specified type.
     * This CompoundValidator may be applied if all validators in the chain and the
     * used formatter my by applied. Disabled validators must by applyable, too!
     * 
     * @param type a type constant defined in {@link at.spardat.enterprise.util.Types Types}.
     */
	public boolean mayBeAppliedTo(byte type) {
		for(Iterator it = chain.iterator();it.hasNext();) {
			if(!((IFmt)it.next()).mayBeAppliedTo(type)) return false;
		}
		if(formatter!=null) return formatter.mayBeAppliedTo(type);
		else return true;
	}

	
    /**
     * Examines whether a given character may be part of an external representation.
     * It must be a legal external char to all enabled validators in the chain.
     * 
     * @param char1 the character to check
     * @return boolean if the character may be part of an external rep.
     */ 
	public boolean isLegalExternalChar(char char1) {
		for(int i=0;i<chain.size();i++) {
			if(switches.get(i)) {
				if(!((IFmt)chain.get(i)).isLegalExternalChar(char1)) {
					return false;
				}
			}
		}
		return true;
	}

    /**
     * Determines if the given string is a legal internal representation.
     * It must be a legal internal char to all enabled validators in the chain.
     * 
     * @param internal the internal encoding.
     * @return whether the provided string follows a legal internal encoding.
     */ 
	public boolean isLegalInternal(String internal) {
		for(int i=0;i<chain.size();i++) {
			if(switches.get(i)) {
				if(!((IFmt)chain.get(i)).isLegalInternal(internal)) {
					return false;
				}
			}
		}
		return true;
	}

    /**
     * Returns the maximum length an external string representation may have.
     * This is the minimum of the maximum lenghts of all enabled validators of the chain.
     * 
     * @return the maximum length or -1 if there is no known maximum.
     */     
	public int maxLenOfExternal() {
		int result=Integer.MAX_VALUE;
		for(int i=0;i<chain.size();i++) {
			if(switches.get(i)) {
				result=Math.min(result, ((IFmt)chain.get(i)).maxLenOfExternal());
			}
		}
		return result;
	}

	/**
     * Returns true, if this formatter only accepts non-empty input.
     * This happens if this CompoundValidator itself is mandatory or
     * any of the enabled validators in the chain is mandatory.
	 */
	public boolean isMandatory() {
		if(super.isMandatory()) return true;
		for(int i=0;i<chain.size();i++) {
			if(switches.get(i)) {
				if(((IFmt)chain.get(i)).isMandatory()) return true;
			}
		}
		return false;
	}

    /**
     * Returns a clone of this.
     * The clone is a deep copy.
     */
	public Object clone() {
		CompoundValidator result = (CompoundValidator) super.clone();
		result.chain=new ArrayList(chain.size());
		if(chain!=null) {
			for(Iterator it=chain.iterator();it.hasNext();) {
				result.chain.add(((IFmt)it.next()).clone());
			}
		}
		if(switches!=null) result.switches=(BitSet) switches.clone();
		if(formatter!=null) result.formatter=(IFmt) formatter.clone();
		return result;
	}

	/**
	 * Clears the validation chain.
	 * Attention: you must add at least one IFmt to the chain
	 * before this CompoundValidator can be used again. 
	 */
	public void clearChain() {
		chain.clear();
		switches=new BitSet();
	}
	
	/**
	 * Replaces the existing validation chain with the new
	 * chain given in the parameter. The list must contain only 
	 * elements of typ IFmt.
	 * @param chain of IFmts to use for validations
	 */
	public void setChain(List chain) {
		this.chain=new ArrayList(chain);
		switches=new BitSet(chain.size());
		switches.set(0,chain.size());
	}
	
	/**
	 * Returns the validation chain.
	 */
	public List getChain() {
		return new ArrayList(chain);
	}
	
	/**
	 * Appends all the IFmts in the parameter to the existing validation chain.
	 */
	public void addToChain(List chain) {
		int oldsize = chain.size();
		this.chain.addAll(chain);
		switches.set(oldsize,chain.size());
	}
	
	/**
	 * Adds the given IFmt to the validation chain.
	 * The new IFmt is added at the end of the chain.
	 * @param fmt to add
	 */
	public void addToChain(IFmt fmt) {
		chain.add(fmt);
		switches.set(chain.size()-1);
	}
	
	/**
	 * Inserts the given IFmt to the validation into the 
	 * validation chain at the given index.
	 */
	public void addToChain(int index,IFmt fmt) {
		chain.add(index, fmt);
		switches=insertBit(switches, index, true);
	}
	
	/**
	 * Removes the given IFmt from the validation chain.
	 * @param fmt to remove
	 */
	public void removeFromChain(IFmt fmt) {
		switches=removeBit(switches,chain.indexOf(fmt));
		chain.remove(fmt);
	}
	
	/**
	 * Removes the IFmt at the given index from the validation chain.
	 * @param index of the IFmt to remove
	 */
	public void removeFromChain(int index) {
		chain.remove(index);
		switches=removeBit(switches,index);
	}

	/**
	 * Returns the index of the given IFmt in the validation chain.
	 * @return the index or -1 if not found.
	 */
	public int indexOf(IFmt fmt) {
		return chain.indexOf(fmt);
	}
	
	/**
	 * Explicitly set the IFmt used for formatting purposes.
	 */
	public void setFormatter(IFmt fmt) {
		formatter=fmt;
	}
	
	/**
	 * Returns the formatter set with {@link #setFormatter(IFmt)}.
	 * If no formatter was explicitly set null is returned.
	 */
	public IFmt getFormatter() {
		return formatter;
	}
	
	/**
	 * Returns the formatter effectively used for formating.
	 * This is either the formatter explicitly set with {@link #setFormatter(IFmt)}
	 * or the first validator in the chain if no formatter was explicitly set.
	 */
	public IFmt getFmtUsedForFormatting() {
		if(formatter!=null) return formatter;
		else return (IFmt)chain.get(0);
	}
	
	/**
	 * Set the enabled state of the validator in the chain at the given index.
	 * @param index of the validator to enable/disable
	 * @param enabled if true enables, if false disables
	 * @return true if the enabled state is now different than before
	 */
	public boolean setEnabled(int index,boolean enabled) {
		boolean changed = switches.get(index) ^ enabled;
		switches.set(index,enabled);
		return changed;
	}
	
	/**
	 * Returns if the validator at the given index in the chain is enabled.
	 */
	public boolean isEnabled(int index) {
		return switches.get(index);
	}
	
	/**
	 * Returns an new BitSet with the given bit inserted at the given index.
	 * All bits after the given index are shifted.
	 */
	private BitSet insertBit(BitSet bitSet,int index,boolean bit) {
		BitSet result = new BitSet(bitSet.size());
		for(int i=0;i<index;i++) result.set(i, bitSet.get(i));
		result.set(index,bit);
		for(int i=index;i<bitSet.size();i++) result.set(i+1, bitSet.get(i));
		return result;
	}

	/**
	 * Returns an new BitSet with one bit removed at the given index.
	 * All bits after the given index are shifted.
	 */
	private BitSet removeBit(BitSet bitSet,int index) {
		BitSet result = new BitSet(bitSet.size());
		for(int i=0;i<index;i++) result.set(i, bitSet.get(i));
		for(int i=index+1;i<bitSet.size();i++) result.set(i-1, bitSet.get(i));
		return result;
	}

}
